"""
A module for instantiating different types of Zeek packages.
"""
import abc
import configparser
import re
import os
import shutil

import semantic_version as semver
import git

from . import (
    __version__,
    LOG,
)
from .package import (
    METADATA_FILENAME,
    name_from_path,
)
from ._util import (
    delete_path,
    git_checkout,
    git_clone,
    git_default_branch,
    git_pull,
    git_version_tags,
    git_remote_urls,
    load_source,
    make_dir,
)

API_VERSION = '1.1.0'

class Error(Exception):
    """Base class for any template-related errors."""

class InputError(Error):
    """Something's amiss in the input arguments for a package."""

class OutputError(Error):
    """Something's going wrong while producing template output."""

class LoadError(Error):
    """Something's going wrong while retrieving a template."""

class GitError(LoadError):
    """There's git trouble while producing template output."""


class Template():
    """Base class for any template.

    Templates need to derive from this class in their toplevel
    __init__.py. Instances of this class pull together the components
    in a given template and capture their parameterization.
    """
    @staticmethod
    def load(config, template, version=None):
        """Template loader.

        This function instantiates a zeekpkg.template.Template
        from a template source present either locally on disk
        or provided as a repo URL.

        It first uses the template's __init__.py to bootstrap a module
        on the fly, then instantiates the zeekpkg.template.Template
        derivative that must be present in it.

        Args:
            config (configparser.ConfigParser): a zkg configuration

            template (str): template source repo, as directory or git URL

            version (str): if provided, a specific version tag to use.
                Ignored when "template" is a local directory. Otherwise,
                the same logic applies as with packages: the most recent
                version tag is picked, and if no version tags are available,
                the default branch.

        Returns:
            zeekpkg.template.Template derivative

        Raises:
            zeekpkg.template.GitError: git hit a problem during cloning/checkout

            zeekpkg.template.LoadError: the template Python code does not load cleanly
        """
        repo = None
        if os.path.isdir(template):
            # We are loading a template from disk. This can be a git
            # repo, in which case we use it as-is. Version requests do
            # not apply. This mirrors the behavior for locally cloned
            # package sources that zkg installs.
            if version is not None:
                LOG.warning('ignoring version request "%s" on local template', version)
            try:
                repo = git.Repo(template)
                if not repo.is_dirty():
                    version = repo.head.ref.commit.hexsha[:8]
            except git.exc.InvalidGitRepositoryError:
                pass
            templatedir = template
        else:
            # We're loading from a git URL. We'll maintain it in the
            # zkg state folder's clone space and support version
            # requests.
            template_clonedir = os.path.join(
                config.get('paths', 'state_dir'), 'clones', 'template')
            templatedir = os.path.join(template_clonedir, name_from_path(template))
            make_dir(template_clonedir)

            try:
                if os.path.isdir(templatedir):
                    # A repo of the requested name is already cloned locally.
                    repo = git.Repo(templatedir)
                    # If the requested URL is not the (only) remote of
                    # this repo, treat it like a different repo and
                    # clone anew. (It could be a fork of the same
                    # repo, for example, or simply a naming
                    # collision.)
                    cur_remote_urls = set()
                    for remote in repo.remotes:
                        cur_remote_urls |= set(remote.urls)
                    if len(cur_remote_urls) == 1 and template in cur_remote_urls:
                        repo.git.fetch('-f', '--recurse-submodules=yes', '--tags')
                    else:
                        delete_path(templatedir)
                        repo = None
                if repo is None:
                    repo = git_clone(template, templatedir)
            except git.exc.GitCommandError as error:
                msg = 'failed to update template "{}": {}'.format(template, error)
                LOG.error(msg)
                raise GitError(msg) from error

            if version is None:
                version_tags = git_version_tags(repo)

                if len(version_tags):
                    version = version_tags[-1]
                else:
                    version = git_default_branch(repo)

            try:
                git_checkout(repo, version)
            except git.exc.GitCommandError as error:
                msg = 'failed to checkout branch/version "{}" of template {}: {}'.format(
                    version, template, error)
                LOG.warn(msg)
                raise GitError(msg) from error

            try:
                # If we're on a branch, pull in latest updates.
                # Pulling fails when on a tag/commit. Accessing the
                # following raises a TypeError when we're not on a
                # branch.
                _ = repo.active_branch
                git_pull(repo)
            except TypeError:
                pass # Not on a branch, do nothing
            except git.exc.GitCommandError as error:
                msg = 'failed to update branch "{}" of template {}: {}'.format(
                    version, template, error)
                LOG.warning(msg)
                raise GitError(msg) from error

        try:
            mod = load_source(os.path.join(templatedir, '__init__.py'))
        except Exception as error:
            msg = 'failed to load template "{}": {}'.format(template, error)
            LOG.exception(msg)
            raise LoadError(msg) from error

        if not hasattr(mod, 'TEMPLATE_API_VERSION'):
            msg = 'template{} does not indicate its API version'.format(
                ' version ' + version if version else '')
            LOG.error(msg)
            raise LoadError(msg)

        # The above guards against absence of TEMPLATE_API_VERSION, so
        # appease pylint for the rest of this function while we access
        # it.
        # pylint: disable=no-member

        try:
            is_compat = Template.is_api_compatible(mod.TEMPLATE_API_VERSION)
        except ValueError:
            raise LoadError('API version string "{}" is invalid'.format(
                mod.TEMPLATE_API_VERSION))

        if not is_compat:
            msg = 'template{} API version is incompatible with zkg ({} vs {})'.format(
                ' version ' + version if version else '',
                mod.TEMPLATE_API_VERSION, API_VERSION)
            LOG.error(msg)
            raise LoadError(msg)

        return mod.Template(templatedir, mod.TEMPLATE_API_VERSION, version, repo)

    @staticmethod
    def is_api_compatible(tmpl_ver):
        """Validate template API compatibility.

        Given a semver string describing the API version for which a
        template was written, verifies that we are compatible with it
        according to semantic versioning rules:

        MAJOR version changes when we make incompatible API changes
        MINOR version changes when you add backwards-compatible functionality
        PATCH version changes when you make backwards-compatible bug fixes.

        Returns:
            bool indicating whether template is comatible.

        Raises:
            ValueError when given version isn't semver-parseable
        """
        tmpl_sv = semver.Version(tmpl_ver)
        api_sv = semver.Version(API_VERSION)

        # A different major version is incompatible by definition
        if tmpl_sv.major != api_sv.major:
            return False

        # Minor version of template can be no larger than ours.
        if tmpl_sv.minor > api_sv.minor:
            return False

        # Patch level does not matter. If ours is less than the
        # template's we're buggy, but the difference doesn't affect API.
        return True

    def __init__(self, templatedir, api_version, version=None, repo=None):
        """Creates a template.

        Template objects start from a local directory, and potentially
        have a version specified. They support the definition and
        lookup of parameters required during the creation of package
        instances from the template. They derive these parameters from
        user variables the template provides and that zkg prompts for
        upon instantiation.

        Args:
            templatedir (str): path to template sources on disk

            api_version (str): API version targeted by the template
                (via TEMPLATE_API_VERSION string)

            version (str): version string of this instance (optional)

            repo (git.Repo): git repo if this template has one (optional)
        """
        self._templatedir = templatedir
        self._api_version = api_version
        self._version = version
        self._repo = repo
        self._params = {} # str -> str, set via self.define_param()
        self._user_vars = []

    def define_user_vars(self):
        """Defines the full set of user vars supported by this template.

        This function defines the complete set of user vars supported
        by the template content. Instances of zeekpkg.template.Package
        and zeekpkg.template.Feature declare which of these user vars
        they require by implementing the needed_user_vars() method,
        returning the names of those variables.

        The default implementation declares no user variables.

        Returns:
            list of zeekpkg.uservar.Uservar instances
        """
        return []

    def apply_user_vars(self, user_vars):
        """Apply the user variables to this template.

        Override this by invoking self.define_param() as needed to create
        template parameters based on the provided user vars. The
        relationship of user vars to template parameters is up to
        you. They can be a 1:1 mapping, you can derive additional
        parameters from a single user var (e.g. to accommodate string
        suffixes), or you can use a combination of user vars to define
        a resulting parameter.

        Args:
            user_vars (list of zeekpkg.uservar.UserVar): input values for the template.
        """

    def package(self):
        """Provides a package template to instantiate.

        If the template provides a Zeek package, return a Package
        instance from this method.

        Returns:
            zeekpkg.template.Package instance
        """
        return None

    def features(self): # pylint: disable=no-self-use
        """Provides any additional features templates supported.

        If the template provides extra features, return each as an
        instance of zeekpkg.template.Feature instance in a list. By
        default, a template offers no features.

        Returns:
            list of zeekpkg.template.Feature instances
        """
        return []

    def templatedir(self):
        """Returns the path to the template's source tree on disk."""
        return self._templatedir

    def name(self):
        """A name for this template, derived from the repo URL."""
        return name_from_path(self._templatedir)

    def api_version(self):
        """The template API version string declared in this instance's module."""
        return self._api_version

    def version(self):
        """A version string for the template.

        This can be a git tag, branch, commit hash, or None if we're
        using the latest version on the default branch.
        """
        return self._version

    def has_repo(self):
        """Returns True if this template has a git repository, False otherwise."""
        return self._repo is not None

    def version_branch(self):
        """Name of the branch the template is on, if any.

        Returns branch name if this template version is a branch HEAD, None
        otherwise (i.e. when it's a specific commit or tag, or we have no
        repository).
        """
        try:
            # The following raises a TypeError when not on a branch
            if self._repo and self._repo.active_branch:
                return self._repo.active_branch.name
        except TypeError:
            pass # Not on a branch

        return None

    def version_sha(self):
        """The git commit hash for this template's version.

        Returns None when this template got instantiated without a git repo,
        otherwise a string with the full hash value in ASCII.
        """
        try:
            if self._repo:
                return self._repo.head.ref.commit.hexsha
        except:
            pass

        return None

    def define_param(self, name, val):
        """Defines a parameter of the given name and value."""
        self._params[name] = val

    def lookup_param(self, name, default=''):
        """Looks up a parameter, falling back to the given default."""
        return self._params.get(name, default)

    def params(self):
        """Returns current str->str template parameter dict."""
        return self._params

    def info(self):
        """Returns a dict capturing information about this template

        This is usable when rendered as JSON, and also serves as the
        input for our human-readable template information.
        """
        # In the future a template may not provide a full package,
        # only features overlaid in combination with another template.
        res = {
            'api_version': self._api_version,
            'provides_package': False,
        }

        # XXX we should revisit the reported 'origin' value in
        # API 2.0.0 -- the the ad-hoc strings are less helpful
        # than simply providing the key only when there's an
        # actual origin.

        if self._repo is not None:
            try:
                remotes = git_remote_urls(self._repo)
                res['origin'] = remotes['origin']
            except KeyError:
                res['origin'] = 'unavailable'
            res['versions'] = git_version_tags(self._repo)
            res['has_repo'] = True
        else:
            res['origin'] = 'not a git repository'
            res['versions'] = []
            res['has_repo'] = False

        pkg = self.package() # pylint: disable=assignment-from-none
        uvars = self.define_user_vars()
        feature_names = []
        res['user_vars'] = {}

        for uvar in uvars:
            res['user_vars'][uvar.name()] = {
                'description': uvar.desc(),
                'default': uvar.default(),
                'used_by': [],
            }

        if pkg is not None:
            res['provides_package'] = True
            for uvar_name in pkg.needed_user_vars():
                try:
                    res['user_vars'][uvar_name]['used_by'].append('package')
                except KeyError:
                    LOG.warning('Package requires undefined user var "%s", skipping', uvar_name)

        for feature in self.features():
            feature_names.append(feature.name())
            for uvar_name in feature.needed_user_vars():
                try:
                    res['user_vars'][uvar_name]['used_by'].append(feature.name())
                except KeyError:
                    LOG.warning('Feature "%s" requires undefined user var "%s"',
                                feature.name(), uvar_name)

        res['features'] = sorted(feature_names)
        return res

    def _set_user_vars(self, user_vars):
        """Provides resolved user vars for the template. Used internally."""
        self._params = {}
        self._user_vars = user_vars
        self.apply_user_vars(user_vars)

    def _get_user_vars(self):
        """Accessor to resolved user vars. Used internally."""
        return self._user_vars


class _Content(metaclass=abc.ABCMeta):
    """Common functionality for all template content."""

    def __init__(self):
        self._features = []
        self._packagedir = None

    @abc.abstractmethod
    def contentdir(self):
        """Subdirectory providing this content in the template tree.

        Returns:
            str: relative path to the content directory
        """
        return None

    def needed_user_vars(self):
        """Returns a list of user vars names required by this content.

        Use this function to declare which of the user vars defined by
        the template (via Template.define_user_vars()) is required by
        this component. By doing this, the user only needs to input
        user vars for template components that actually require them.

        Returns:
            A list of strings identifying the needed user vars.
        """
        return []

    def add_feature(self, feature):
        self._features.append(feature)

    def do_validate(self, tmpl):
        """Main driver for validation of a template's configuration.

        zkg calls this internally as part of template validation.
        You'll likely want to focus on validate() for
        customization.
        """
        self.validate(tmpl)
        for feature in self._features:
            feature.validate(tmpl)

    def validate(self, tmpl):
        """Validation of template configuration for this component.

        Override this in your template's code in order to check
        whether the template parameters (available via
        tmpl.lookup_param()) are present and correctly
        formatted. Raise zeekpkg.template.InputError exceptions as
        needed.

        Args:
            tmpl (zeekpkg.template.Template): template context

        Raises:
            zeekpkg.template.InputError when failing validation.

        """

    def do_instantiate(self, tmpl, packagedir, use_force=False):
        """Main driver for instantiating template content.

        zkg calls this internally as part of template instantiation.
        You'll likely want to focus on instantiate() for
        customization.

        Args:
            tmpl (zeekpkg.template.Template): template context

            packagedir (str): output folder for the instantiation

            use_force (bool): whether to overwrite/recreate files as needed

        """
        self._packagedir = packagedir

        self.instantiate(tmpl)

        for feature in self._features:
            feature.do_instantiate(tmpl, packagedir, use_force=use_force)

    def instantiate(self, tmpl):
        """Instantiation of this template component.

        This substitutes parameters in the template material and
        instantiates the result in the output directory.

        Args:
            tmpl (zeekpkg.template.Template): template context
        """
        for orig_file, path_name, file_name, content in self._walk(tmpl):
            if os.path.islink(orig_file):
                self.instantiate_symlink(tmpl, orig_file, path_name, file_name, content)
            else:
                self.instantiate_file(tmpl, orig_file, path_name, file_name, content)

    def instantiate_file(self, tmpl, orig_file, path_name, file_name, content):
        """Instantiate a regular file in the template.

        This gets invoked by instantiate() as it traverses the
        template content. Each invocation sees the content after
        parameter substitution in file/directory names and file
        content.

        Directories get handled implicitly via the path_name of any
        files contained in it.

        This implementation writes the content to the output file,
        overwriting any pre-existing output.

        Args:
            tmpl (zeekpkg.template.Template): template context

            orig_file (str): the absolute input file name, e.g. "/path/to/template/@param@.zeek"

            path_name (str): the output directory inside the --packagedir

            file_name (str): the resulting output file name, e.g. "result.zeek"

            content (str or bytes): the resulting content for the file.

        """
        out_dir = os.path.join(self._packagedir, path_name)
        out_file = os.path.join(out_dir, file_name)
        os.makedirs(out_dir, exist_ok=True)

        try:
            with open(out_file, 'wb') as hdl:
                hdl.write(content)
            shutil.copymode(orig_file, out_file)
        except IOError as error:
            LOG.warning('I/O error while instantiating "%s": %s', out_file, error)

    def instantiate_symlink(self, tmpl, orig_file, path_name, file_name, target):
        """Instantiate a symbolic link in the template.

        This gets invoked by instantiate() as it traverses the
        template content. Each invocation sees the symlink target
        after parameter substitution.

        Directories get handled implicitly via the path_name of any
        files contained in it.

        This implementation deletes existing files and creates the
        symlink in their place.

        Args:
            tmpl (zeekpkg.template.Template): template context

            orig_file (str): the absolute input file name, e.g. "/path/to/template/@param@.zeek"

            path_name (str): the output directory inside the --packagedir

            file_name (str): the resulting output file name, e.g. "result.zeek"

            target (str): the location the symlink points to.
        """
        out_dir = os.path.join(self._packagedir, path_name)
        out_file = os.path.join(out_dir, file_name)
        os.makedirs(out_dir, exist_ok=True)

        try:
            delete_path(out_file)
            os.symlink(target, out_file)
        except OSError as error:
            LOG.warning('OS error while creating symlink "%s": %s',
                        out_file, error)

    def _walk(self, tmpl):
        """Generator for instantiating template content.

        This walks over the template source tree, yielding for every
        file a 4-tuple of the input file name, the output directory,
        the output file name in that directory, and the file's
        content. For symlinks, the content is the symlink target,
        with any applicable parameter subtitutions made.

        Args:
            tmpl (zeekpkg.template.Template): template context
        """
        prefix = os.path.join(tmpl.templatedir(), self.contentdir())
        for root, _, files in os.walk(prefix):
            for fname in files:
                in_file = root + os.sep + fname

                # Substitute directory and file names
                out_path = self._replace(tmpl, root[len(prefix)+1:])
                out_file = self._replace(tmpl, fname)

                if os.path.islink(in_file):
                    out_content = self._replace(tmpl, os.readlink(in_file))
                else:
                    # Substitute file content.
                    try:
                        with open(in_file, 'rb') as hdl:
                            out_content = self._replace(tmpl, hdl.read())
                    except IOError as error:
                        LOG.warning('skipping instantiation of %s: %s', in_file, error)
                        continue

                yield in_file, out_path, out_file, out_content

    def _replace(self, tmpl, content): # pylint: disable=no-self-use
        """Helper for content substitution.

        Args:
            tmpl (zeekpkg.template.Template): template context

            content (str or bytes): unsubstituted template fodder

        Returns:
            str or bytes after parameter substitution.
        """
        for name, val in tmpl.params().items():
            pat = '@' + name + '@'
            if not isinstance(content, str):
                pat = bytes(pat, 'ascii')
                val = bytes(val, 'ascii')
            content = re.sub(pat, val, content, flags=re.IGNORECASE)

        return content


class Package(_Content):
    """Template content for a Zeek package.

    This class fills in package-specific functionality but it still
    abstract. At a minimum, your template's Package derivative needs
    to implement contentdir().
    """
    def do_instantiate(self, tmpl, packagedir, use_force=False):
        self._prepare_packagedir(packagedir)
        super().do_instantiate(tmpl, packagedir, use_force)
        self._update_metadata(tmpl)
        self._git_init(tmpl)

    def _prepare_packagedir(self, packagedir):
        os.makedirs(packagedir, exist_ok=True)

    def _update_metadata(self, tmpl):
        """Updates the package's zkg.meta with template information.

        This information allows re-running template instantiation with
        identical inputs at a later time.
        """
        config = configparser.ConfigParser(delimiters='=')
        config.optionxform = str
        manifest_file = os.path.join(self._packagedir, METADATA_FILENAME)

        # Best-effort: if the template populated the file, adopt the
        # content, otherwise create with just our metadata.
        config.read(manifest_file)

        section = 'template'
        config.remove_section(section)
        config.add_section(section)
        config.set(section, 'source', tmpl.name())

        if tmpl.has_repo():
            tmplinfo = tmpl.info()
            if tmplinfo['origin'] != 'unavailable':
                config.set(section, 'source', tmplinfo['origin'])

        if tmpl.version():
            # If we're on a branch, disambiguate the version by also mentioning
            # the exact commit.
            if tmpl.version_branch():
                config.set(section, 'version', tmpl.version_branch())
                config.set(section, 'commit', tmpl.version_sha()[:8])
            else:
                config.set(section, 'version', tmpl.version())
        else:
            config.set(section, 'version', tmpl.version() or 'unversioned')

        config.set(section, 'zkg_version', __version__)

        if self._features:
            val = ','.join(sorted([f.name() for f in self._features]))
            config.set(section, 'features', val)

        section = 'template_vars'
        config.remove_section(section)
        config.add_section(section)

        for uvar in tmpl._get_user_vars(): # pylint: disable=protected-access
            if uvar.val() is not None:
                config.set(section, uvar.name(), uvar.val())

        with open(manifest_file, 'w') as hdl:
            config.write(hdl)

    def _git_init(self, tmpl):
        """Initialize git repo and commit instantiated content."""
        repo = git.Repo.init(self._packagedir)
        for fname in repo.untracked_files:
            repo.index.add(fname)

        features_info = ''
        if self._features:
            names = sorted(['"' + f.name() + '"' for f in self._features])
            if len(names) == 1:
                features_info = ', with feature {}'.format(names[0])
            else:
                features_info = ', with features '
                features_info += ', '.join(names[:-1])
                features_info += ' and ' + names[-1]

        ver_info = tmpl.version()
        ver_sha = tmpl.version_sha()

        if ver_info is None:
            if ver_sha:
                ver_info = 'version ' + ver_sha[:8]
            else:
                ver_info = 'no versioning'
        else:
            ver_info = 'version ' + ver_info
            if ver_sha:
                ver_info += ' (' + ver_sha[:8] + ')'

        repo.index.commit("""Initial commit.

zkg {} created this package from template "{}"
using {}{}.""".format(__version__, tmpl.name(), ver_info, features_info))


class Feature(_Content):
    """Features overlay additional functionality onto a package.

    This class fills in feature-specific functionality but it still
    abstract. At a minimum, your template's Feature derivative needs
    to implement contentdir().
    """
    def name(self):
        """A name for this feature. Defaults to its content directory."""
        return self.contentdir() or 'unnamed'



"""
A module containing the definition of a "package source": a git repository
containing a collection of :file:`zkg.index` (or legacy :file:`bro-pkg.index`)
files.  These are simple INI files that can describe many Zeek packages.  Each
section of the file names a Zeek package along with the git URL where it is
located and metadata tags that help classify/describe it.
"""

import configparser
import git
import os
import shutil

from . import LOG
from .package import (
    name_from_path,
    Package
)
from ._util import (
    git_default_branch,
    git_checkout,
    git_clone
)

#: The name of package index files.
INDEX_FILENAME = 'zkg.index'
LEGACY_INDEX_FILENAME = 'bro-pkg.index'
#: The name of the package source file where package metadata gets aggregated.
AGGREGATE_DATA_FILE = 'aggregate.meta'


class Source(object):
    """A Zeek package source.

    This class contains properties of a package source like its name, remote git
    URL, and local git clone.

    Attributes:
        name (str): The name of the source as given by a config file key
            in it's ``[sources]`` section.

        git_url (str): The git URL of the package source.

        clone (git.Repo): The local git clone of the package source.
    """

    def __init__(self, name, clone_path, git_url, version=None):
        """Create a package source.

        Raises:
            git.exc.GitCommandError: if the git repo is invalid
            OSError: if the git repo is invalid and can't be re-initialized
        """
        git_url = os.path.expanduser(git_url)
        self.name = name
        self.git_url = git_url
        self.clone = None

        try:
            self.clone = git.Repo(clone_path)
        except git.exc.NoSuchPathError:
            LOG.debug('creating source clone of "%s" at %s', name, clone_path)
            self.clone = git_clone(git_url, clone_path, shallow=True)
        except git.exc.InvalidGitRepositoryError:
            LOG.debug('deleting invalid source clone of "%s" at %s',
                      name, clone_path)
            shutil.rmtree(clone_path)
            self.clone = git_clone(git_url, clone_path, shallow=True)
        else:
            LOG.debug('found source clone of "%s" at %s', name, clone_path)
            old_url = self.clone.git.config('--local', '--get',
                                            'remote.origin.url')

            if git_url != old_url:
                LOG.debug(
                    'url of source "%s" changed from %s to %s, reclone at %s',
                    name, old_url, git_url, clone_path)
                shutil.rmtree(clone_path)
                self.clone = git_clone(git_url, clone_path, shallow=True)

        git_checkout(self.clone, version or git_default_branch(self.clone))

    def __str__(self):
        return self.git_url

    def __repr__(self):
        return self.git_url

    def package_index_files(self):
        """Return a list of paths to package index files in the source."""
        rval = []
        visited_dirs = set()

        for root, dirs, files in os.walk(self.clone.working_dir,
                                         followlinks=True):
            stat = os.stat(root)
            visited_dirs.add((stat.st_dev, stat.st_ino))
            dirs_to_visit_next = []

            for d in dirs:
                stat = os.stat(os.path.join(root, d))

                if (stat.st_dev, stat.st_ino) not in visited_dirs:
                    dirs_to_visit_next.append(d)

            dirs[:] = dirs_to_visit_next

            try:
                dirs.remove('.git')
            except ValueError:
                pass

            for filename in files:
                if filename == INDEX_FILENAME or filename == LEGACY_INDEX_FILENAME:
                    rval.append(os.path.join(root, filename))

        return sorted(rval)

    def packages(self):
        """Return a list of :class:`.package.Package` in the source."""
        rval = []
        # Use raw parser so no value interpolation takes place.
        parser = configparser.RawConfigParser()
        aggregate_file = os.path.join(
            self.clone.working_dir, AGGREGATE_DATA_FILE)
        parser.read(aggregate_file)

        for index_file in self.package_index_files():
            relative_path = index_file[len(self.clone.working_dir) + 1:]
            directory = os.path.dirname(relative_path)
            lines = []

            with open(index_file) as f:
                lines = [line.rstrip('\n') for line in f]

            for url in lines:
                pkg_name = name_from_path(url)
                agg_key = os.path.join(directory, pkg_name)
                metadata = {}

                if parser.has_section(agg_key):
                    metadata = {key: value for key,
                                value in parser.items(agg_key)}

                package = Package(git_url=url, source=self.name,
                                  directory=directory, metadata=metadata)
                rval.append(package)

        return rval



from argparse import ArgumentParser
import os

from docutils import nodes
from docutils.statemachine import StringList
from docutils.parsers.rst.directives import flag, unchanged

try:
    # Removed as of Sphinx 1.7
    from sphinx.util.compat import Directive
except ImportError as err:
    from docutils.parsers.rst import Directive

from sphinx.util.nodes import nested_parse_with_titles

from sphinxarg.parser import parse_parser, parser_navigate


def map_nested_definitions(nested_content):
    if nested_content is None:
        raise Exception('Nested content should be iterable, not null')
    # build definition dictionary
    definitions = {}
    for item in nested_content:
        if not isinstance(item, nodes.definition_list):
            continue
        for subitem in item:
            if not isinstance(subitem, nodes.definition_list_item):
                continue
            if not len(subitem.children) > 0:
                continue
            classifier = '@after'
            idx = subitem.first_child_matching_class(nodes.classifier)
            if idx is not None:
                ci = subitem[idx]
                if len(ci.children) > 0:
                    classifier = ci.children[0].astext()
            if classifier is not None and classifier not in (
                    '@replace', '@before', '@after'):
                raise Exception('Unknown classifier: %s' % classifier)
            idx = subitem.first_child_matching_class(nodes.term)
            if idx is not None:
                ch = subitem[idx]
                if len(ch.children) > 0:
                    term = ch.children[0].astext()
                    idx = subitem.first_child_matching_class(nodes.definition)
                    if idx is not None:
                        def_node = subitem[idx]
                        def_node.attributes['classifier'] = classifier
                        definitions[term] = def_node
    return definitions


def print_arg_list(data, nested_content):
    definitions = map_nested_definitions(nested_content)
    items = []
    if 'args' in data:
        for arg in data['args']:
            my_def = [nodes.paragraph(text=arg['help'])] if arg['help'] else []
            name = arg['name']
            my_def = apply_definition(definitions, my_def, name)
            if len(my_def) == 0:
                my_def.append(nodes.paragraph(text='Undocumented'))
            if 'choices' in arg:
                my_def.append(nodes.paragraph(
                    text=('Possible choices: %s' % ', '.join([str(c) for c in arg['choices']]))))
            argname = name

            if arg['metavar']:
                argname = arg['metavar']

            items.append(
                nodes.option_list_item(
                    '', nodes.option_group('',
                          nodes.option('', nodes.option_string(text=argname))),
                    nodes.description('', *my_def)))
    return nodes.option_list('', *items) if items else None


def print_opt_list(data, nested_content):
    definitions = map_nested_definitions(nested_content)
    items = []
    if 'options' in data:
        for opt in data['options']:
            names = []
            my_def = [nodes.paragraph(text=opt['help'])] if opt['help'] else []
            for name in opt['name']:
                option_declaration = [nodes.option_string(text=name)]
                if opt['default'] is not None \
                        and opt['default'] != '==SUPPRESS==':
                    option_declaration += nodes.option_argument(
                        '', text='=' + str(opt['default']))
                names.append(nodes.option('', *option_declaration))
                my_def = apply_definition(definitions, my_def, name)
            if len(my_def) == 0:
                my_def.append(nodes.paragraph(text='Undocumented'))
            if 'choices' in opt:
                my_def.append(nodes.paragraph(
                    text=('Possible choices: %s' % ', '.join([str(c) for c in opt['choices']]))))
            items.append(
                nodes.option_list_item(
                    '', nodes.option_group('', *names),
                    nodes.description('', *my_def)))
    return nodes.option_list('', *items) if items else None


def print_command_args_and_opts(arg_list, opt_list, sub_list=None):
    items = []
    if arg_list:
        items.append(nodes.definition_list_item(
            '', nodes.term(text='Positional arguments:'),
            nodes.definition('', arg_list)))
    if opt_list:
        items.append(nodes.definition_list_item(
            '', nodes.term(text='Options:'),
            nodes.definition('', opt_list)))
    if sub_list and len(sub_list):
        items.append(nodes.definition_list_item(
            '', nodes.term(text='Sub-commands:'),
            nodes.definition('', sub_list)))
    return nodes.definition_list('', *items)


def apply_definition(definitions, my_def, name):
    if name in definitions:
        definition = definitions[name]
        classifier = definition['classifier']
        if classifier == '@replace':
            return definition.children
        if classifier == '@after':
            return my_def + definition.children
        if classifier == '@before':
            return definition.children + my_def
        raise Exception('Unknown classifier: %s' % classifier)
    return my_def


def print_subcommand_list(data, nested_content):
    definitions = map_nested_definitions(nested_content)
    items = []
    if 'children' in data:
        for child in data['children']:
            my_def = [nodes.paragraph(
                text=child['help'])] if child['help'] else []
            name = child['name']
            my_def = apply_definition(definitions, my_def, name)
            if len(my_def) == 0:
                my_def.append(nodes.paragraph(text='Undocumented'))
            if 'description' in child:
                my_def.append(nodes.paragraph(text=child['description']))
            my_def.append(nodes.literal_block(text=child['usage']))
            my_def.append(print_command_args_and_opts(
                print_arg_list(child, nested_content),
                print_opt_list(child, nested_content),
                print_subcommand_list(child, nested_content)
            ))
            items.append(
                nodes.definition_list_item(
                    '',
                    nodes.term('', '', nodes.strong(text=name)),
                    nodes.definition('', *my_def)
                )
            )
    return nodes.definition_list('', *items)


class ArgParseDirective(Directive):
    has_content = True
    option_spec = dict(module=unchanged, func=unchanged, ref=unchanged,
                       prog=unchanged, path=unchanged, nodefault=flag,
                       manpage=unchanged, nosubcommands=unchanged, passparser=flag)

    def _construct_manpage_specific_structure(self, parser_info):
        """
        Construct a typical man page consisting of the following elements:
            NAME (automatically generated, out of our control)
            SYNOPSIS
            DESCRIPTION
            OPTIONS
            FILES
            SEE ALSO
            BUGS
        """
        # SYNOPSIS section
        synopsis_section = nodes.section(
            '',
            nodes.title(text='Synopsis'),
            nodes.literal_block(text=parser_info["bare_usage"]),
            ids=['synopsis-section'])
        # DESCRIPTION section
        description_section = nodes.section(
            '',
            nodes.title(text='Description'),
            nodes.paragraph(text=parser_info.get(
                'description', parser_info.get(
                    'help', "undocumented").capitalize())),
            ids=['description-section'])
        nested_parse_with_titles(
            self.state, self.content, description_section)
        if parser_info.get('epilog'):
            # TODO: do whatever sphinx does to understand ReST inside
            # docstrings magically imported from other places. The nested
            # parse method invoked above seem to be able to do this but
            # I haven't found a way to do it for arbitrary text
            description_section += nodes.paragraph(
                text=parser_info['epilog'])
        # OPTIONS section
        options_section = nodes.section(
            '',
            nodes.title(text='Options'),
            ids=['options-section'])
        if 'args' in parser_info:
            options_section += nodes.paragraph()
            options_section += nodes.subtitle(text='Positional arguments:')
            options_section += self._format_positional_arguments(parser_info)
        if 'options' in parser_info:
            options_section += nodes.paragraph()
            options_section += nodes.subtitle(text='Optional arguments:')
            options_section += self._format_optional_arguments(parser_info)
        items = [
            # NOTE: we cannot generate NAME ourselves. It is generated by
            # docutils.writers.manpage
            synopsis_section,
            description_section,
            # TODO: files
            # TODO: see also
            # TODO: bugs
        ]
        if len(options_section.children) > 1:
            items.append(options_section)
        if 'nosubcommands' not in self.options:
            # SUBCOMMANDS section (non-standard)
            subcommands_section = nodes.section(
                '',
                nodes.title(text='Sub-Commands'),
                ids=['subcommands-section'])
            if 'children' in parser_info:
                subcommands_section += self._format_subcommands(parser_info)
            if len(subcommands_section) > 1:
                items.append(subcommands_section)
        if os.getenv("INCLUDE_DEBUG_SECTION"):
            import json
            # DEBUG section (non-standard)
            debug_section = nodes.section(
                '',
                nodes.title(text="Argparse + Sphinx Debugging"),
                nodes.literal_block(text=json.dumps(parser_info, indent='  ')),
                ids=['debug-section'])
            items.append(debug_section)
        return items

    def _format_positional_arguments(self, parser_info):
        assert 'args' in parser_info
        items = []
        for arg in parser_info['args']:
            arg_items = []
            if arg['help']:
                arg_items.append(nodes.paragraph(text=arg['help']))
            else:
                arg_items.append(nodes.paragraph(text='Undocumented'))
            if 'choices' in arg:
                arg_items.append(
                    nodes.paragraph(
                        text='Possible choices: ' + ', '.join(arg['choices'])))
            items.append(
                nodes.option_list_item(
                    '',
                    nodes.option_group(
                        '', nodes.option(
                            '', nodes.option_string(text=arg['metavar'])
                        )
                    ),
                    nodes.description('', *arg_items)))
        return nodes.option_list('', *items)

    def _format_optional_arguments(self, parser_info):
        assert 'options' in parser_info
        items = []
        for opt in parser_info['options']:
            names = []
            opt_items = []
            for name in opt['name']:
                option_declaration = [nodes.option_string(text=name)]
                if opt['default'] is not None \
                        and opt['default'] != '==SUPPRESS==':
                    option_declaration += nodes.option_argument(
                        '', text='=' + str(opt['default']))
                names.append(nodes.option('', *option_declaration))
            if opt['help']:
                opt_items.append(nodes.paragraph(text=opt['help']))
            else:
                opt_items.append(nodes.paragraph(text='Undocumented'))
            if 'choices' in opt:
                opt_items.append(
                    nodes.paragraph(
                        text='Possible choices: ' + ', '.join(opt['choices'])))
            items.append(
                nodes.option_list_item(
                    '', nodes.option_group('', *names),
                    nodes.description('', *opt_items)))
        return nodes.option_list('', *items)

    def _format_subcommands(self, parser_info):
        assert 'children' in parser_info
        items = []
        for subcmd in parser_info['children']:
            subcmd_items = []
            if subcmd['help']:
                subcmd_items.append(nodes.paragraph(text=subcmd['help']))
            else:
                subcmd_items.append(nodes.paragraph(text='Undocumented'))
            items.append(
                nodes.definition_list_item(
                    '',
                    nodes.term('', '', nodes.strong(
                        text=subcmd['bare_usage'])),
                    nodes.definition('', *subcmd_items)))
        return nodes.definition_list('', *items)

    def _nested_parse_paragraph(self, text):
        content = nodes.paragraph()
        self.state.nested_parse(StringList(text.split("\n")), 0, content)
        return content

    def run(self):
        if 'module' in self.options and 'func' in self.options:
            module_name = self.options['module']
            attr_name = self.options['func']
        elif 'ref' in self.options:
            _parts = self.options['ref'].split('.')
            module_name = '.'.join(_parts[0:-1])
            attr_name = _parts[-1]
        else:
            raise self.error(
                ':module: and :func: should be specified, or :ref:')
        mod = __import__(module_name, globals(), locals(), [attr_name])
        file_dependency = mod.__file__

        if file_dependency.endswith('.pyc'):
            file_dependency = file_dependency[:-1]

        env = self.state.document.settings.env

        if not hasattr(env, 'argparse_usages'):
            env.argparse_usages = []

        env.argparse_usages.append({
            'docname': env.docname,
            'lineno': self.lineno,
            'dependency_file': file_dependency,
            'dependency_mtime': os.stat(file_dependency).st_mtime,
        })

        if not hasattr(mod, attr_name):
            raise self.error((
                'Module "%s" has no attribute "%s"\n'
                'Incorrect argparse :module: or :func: values?'
            ) % (module_name, attr_name))
        func = getattr(mod, attr_name)
        if isinstance(func, ArgumentParser):
            parser = func
        elif 'passparser' in self.options:
            parser = ArgumentParser()
            func(parser)
        else:
            parser = func()
        if 'path' not in self.options:
            self.options['path'] = ''
        path = str(self.options['path'])
        if 'prog' in self.options:
            parser.prog = self.options['prog']
        result = parse_parser(
            parser, skip_default_values='nodefault' in self.options)
        result = parser_navigate(result, path)
        if 'manpage' in self.options:
            return self._construct_manpage_specific_structure(result)
        nested_content = nodes.paragraph()
        self.state.nested_parse(
            self.content, self.content_offset, nested_content)
        nested_content = nested_content.children
        items = []
        # add common content between
        for item in nested_content:
            if not isinstance(item, nodes.definition_list):
                items.append(item)
        if 'description' in result:
            items.append(self._nested_parse_paragraph(result['description']))
        items.append(nodes.literal_block(text=result['usage']))

        if 'nosubcommands' in self.options:
            subcommands = None
        else:
            subcommands = print_subcommand_list(result, nested_content)

        items.append(print_command_args_and_opts(
            print_arg_list(result, nested_content),
            print_opt_list(result, nested_content),
            subcommands
        ))
        if 'epilog' in result:
            items.append(self._nested_parse_paragraph(result['epilog']))
        return items


def env_get_outdated_hook(app, env, added, changed, removed):
    from sphinx.util import logging
    logger = logging.getLogger(__name__)

    rval = set()

    if not hasattr(env, 'argparse_usages'):
        return []

    for usage in env.argparse_usages:
        docname = usage['docname']
        dep_file = usage['dependency_file']
        dep_mtime = usage['dependency_mtime']
        current_mtime = os.stat(dep_file).st_mtime

        if current_mtime > dep_mtime and docname not in removed:
            rval.add(docname)

    for docname in rval:
        from sphinx.util.console import blue
        msg = blue('found outdated argparse doc: {0}'.format(docname))
        logger.info(msg)

    return list(rval)


def env_purge_doc_hook(app, env, docname):
    if not hasattr(env, 'argparse_usages'):
        return

    env.argparse_usages = [
        usage for usage in env.argparse_usages if usage['docname'] != docname]


def setup(app):
    app.add_directive('argparse', ArgParseDirective)
    app.connect('env-get-outdated', env_get_outdated_hook)
    app.connect('env-purge-doc', env_purge_doc_hook)



"""
This package defines a Python interface for installing, managing, querying,
and performing other operations on Zeek Packages and Package Sources.
The main entry point is the :class:`Manager <zeekpkg.manager.Manager>` class.

This package provides a logger named ``LOG`` to which logging stream handlers
may be added in order to help log/debug applications.
"""

import logging

__version__ = "2.13.0"
__all__ = ['manager', 'package', 'source', 'template', 'uservar']

LOG = logging.getLogger(__name__)
LOG.addHandler(logging.NullHandler())

from .manager import *
from .package import *
from .source import *
from .uservar import *



import zeekpkg.template
import zeekpkg.uservar

TEMPLATE_API_VERSION = '1.0.0'

class Package(zeekpkg.template.Package):
    def contentdir(self):
        return 'package'

    def needed_user_vars(self):
        return ['name']

    def validate(self, tmpl):
        if not tmpl.lookup_param('name'):
            raise zeekpkg.template.InputError(
                'package requires a name')


class Readme(zeekpkg.template.Feature):
    def contentdir(self):
        return 'readme'

    def needed_user_vars(self):
        return ['readme']


class Template(zeekpkg.template.Template):
    def define_user_vars(self):
        return [
            zeekpkg.uservar.UserVar(
                'name', desc='the name of the package, e.g. "FooBar"'),
            zeekpkg.uservar.UserVar(
                'readme', desc='Content of the README file',
                val='This is a README.'),
        ]

    def apply_user_vars(self, uvars):
        for uvar in uvars:
            if uvar.name() == 'name':
                self.define_param('name', uvar.val())
                self.define_param('module', uvar.val().upper())
            if uvar.name() == 'readme':
                self.define_param('readme', uvar.val())

    def package(self):
        return Package()

    def features(self):
        return [Readme()]



"""
A module with various data structures used for interacting with and querying
the properties and status of Zeek packages.
"""

import os
import re

from .uservar import UserVar
from ._util import find_sentence_end

#: The name of files used by packages to store their metadata.
METADATA_FILENAME = 'zkg.meta'
LEGACY_METADATA_FILENAME = 'bro-pkg.meta'

TRACKING_METHOD_VERSION = 'version'
TRACKING_METHOD_BRANCH = 'branch'
TRACKING_METHOD_COMMIT = 'commit'

PLUGIN_MAGIC_FILE = '__bro_plugin__'
PLUGIN_MAGIC_FILE_DISABLED = '__bro_plugin__.disabled'

def name_from_path(path):
    """Returns the name of a package given a path to its git repository."""
    return canonical_url(path).split('/')[-1]


def canonical_url(path):
    """Returns the url of a package given a path to its git repo."""
    url = path.rstrip('/')

    if url.startswith('.') or url.startswith('/'):
        url = os.path.realpath(url)

    return url


def is_valid_name(name):
    """Returns True if name is a valid package name, else False."""
    if name != name.strip():
        # Reject names with leading/trailing whitespace
        return False

    if name in ("package", "packages"):
        return False

    return True


def aliases(metadata_dict):
    """Return a list of package aliases found in metadata's 'aliases' field."""
    if 'aliases' not in metadata_dict:
        return []

    return re.split(',\s*|\s+', metadata_dict['aliases'])


def tags(metadata_dict):
    """Return a list of tag strings found in the metadata's 'tags' field."""
    if 'tags' not in metadata_dict:
        return []

    return re.split(',\s*', metadata_dict['tags'])


def short_description(metadata_dict):
    """Returns the first sentence of the metadata's 'desciption' field."""
    if 'description' not in metadata_dict:
        return ''

    description = metadata_dict['description']
    lines = description.split('\n')
    rval = ''

    for line in lines:
        line = line.lstrip()
        rval += ' '
        period_idx = find_sentence_end(line)

        if period_idx == -1:
            rval += line
        else:
            rval += line[:period_idx + 1]
            break

    return rval.lstrip()


def user_vars(metadata_dict):
    """Returns a list of (str, str, str) from metadata's 'user_vars' field.

    Each entry in the returned list is a the name of a variable, its value,
    and its description.

    If the 'user_vars' field is not present, an empty list is returned.  If it
    is malformed, then None is returned.
    """
    uvars = UserVar.parse_dict(metadata_dict)

    if uvars is None:
        return None

    return [(uvar.name(), uvar.val(), uvar.desc()) for uvar in uvars]


def dependencies(metadata_dict, field='depends'):
    """Returns a dictionary of (str, str) based on metadata's dependency field.

    The keys indicate the name of a package (shorthand name or full git URL).
    The names 'zeek' or 'zkg' may also be keys that indicate a dependency on a
    particular Zeek or zkg version.

    The values indicate a semantic version requirement.

    If the dependency field is malformed (e.g. number of keys not equal to
    number of values), then None is returned.
    """
    if field not in metadata_dict:
        return dict()

    rval = dict()
    depends = metadata_dict[field]
    parts = depends.split()
    keys = parts[::2]
    values = parts[1::2]

    if len(keys) != len(values):
        return None

    for i, k in enumerate(keys):
        if i < len(values):
            rval[k] = values[i]

    return rval


class InstalledPackage(object):
    """An installed package and its current status.

    Attributes:
        package (:class:`Package`): the installed package

        status (:class:`PackageStatus`): the status of the installed package
    """

    def __init__(self, package, status):
        self.package = package
        self.status = status

    def __lt__(self, other):
        return str(self.package) < str(other.package)


class PackageStatus(object):
    """The status of an installed package.

    This class contains properties of a package related to how the package
    manager will operate on it.

    Attributes:
        is_loaded (bool): whether a package is marked as "loaded".

        is_pinned (bool): whether a package is allowed to be upgraded.

        is_outdated (bool): whether a newer version of the package exists.

        tracking_method (str): either "branch", "version", or "commit" to
            indicate (respectively) whether package upgrades should stick to a
            git branch, use git version tags, or do nothing because the
            package is to always use a specific git commit hash.

        current_version (str): the current version of the installed
            package, which is either a git branch name or a git version tag.

        current_hash (str): the git sha1 hash associated with installed
            package's current version/commit.
    """

    def __init__(self, is_loaded=False, is_pinned=False, is_outdated=False,
                 tracking_method=None, current_version=None, current_hash=None):
        self.is_loaded = is_loaded
        self.is_pinned = is_pinned
        self.is_outdated = is_outdated
        self.tracking_method = tracking_method
        self.current_version = current_version
        self.current_hash = current_hash


class PackageInfo(object):
    """Contains information on an arbitrary package.

    If the package is installed, then its status is also available.

    Attributes:
        package (:class:`Package`): the relevant Zeek package

        status (:class:`PackageStatus`): this attribute is set for installed
            packages

        metadata (dict of str -> str): the contents of the package's
            :file:`zkg.meta` or :file:`bro-pkg.meta`

        versions (list of str): a list of the package's availabe git version
            tags

        metadata_version: the package version that the metadata is from

        version_type: either 'version', 'branch', or 'commit' to
            indicate whether the package info/metadata was taken from a release
            version tag, a branch, or a specific commit hash.

        invalid_reason (str): this attribute is set when there is a problem
            with gathering package information and explains what went wrong.

        metadata_file: the absolute path to the :file:`zkg.meta` or
            :file:`bro-pkg.meta` for this package.  Use this if you'd like to
            parse the metadata yourself. May not be defined, in which case the
            value is None.
    """

    def __init__(self, package=None, status=None, metadata=None, versions=None,
                 metadata_version='', invalid_reason='', version_type='',
                 metadata_file=None, default_branch=None):
        self.package = package
        self.status = status
        self.metadata = {} if metadata is None else metadata
        self.versions = [] if versions is None else versions
        self.metadata_version = metadata_version
        self.version_type = version_type
        self.invalid_reason = invalid_reason
        self.metadata_file = metadata_file
        self.default_branch = default_branch

    def aliases(self):
        """Return a list of package name aliases.

        The canonical one is listed first.
        """
        return aliases(self.metadata)

    def tags(self):
        """Return a list of keyword tags associated with the package.

        This will be the contents of the package's `tags` field."""
        return tags(self.metadata)

    def short_description(self):
        """Return a short description of the package.

        This will be the first sentence of the package's 'description' field."""
        return short_description(self.metadata)

    def dependencies(self, field='depends'):
        """Returns a dictionary of dependency -> version strings.

        The keys indicate the name of a package (shorthand name or full git
        URL).  The names 'zeek' or 'zkg' may also be keys that indicate a
        dependency on a particular Zeek or zkg version.

        The values indicate a semantic version requirement.

        If the dependency field is malformed (e.g. number of keys not equal to
        number of values), then None is returned.
        """
        return dependencies(self.metadata, field)

    def user_vars(self):
        """Returns a list of user variables parsed from metadata's 'user_vars' field.

        If the 'user_vars' field is not present, an empty list is returned.  If
        it is malformed, then None is returned.

        Returns:
            list of zeekpkg.uservar.UserVar, or None on error
        """
        return UserVar.parse_dict(self.metadata)

    def best_version(self):
        """Returns the best/latest version of the package that is available.

        If the package has any git release tags, this returns the highest one,
        else it returns the default branch like 'main' or 'master'.
        """
        if self.versions:
            return self.versions[-1]

        return self.default_branch


class Package(object):
    """A Zeek package.

    This class contains properties of a package that are defined by the package
    git repository itself and the package source it came from.

    Attributes:
        git_url (str): the git URL which uniquely identifies where the
            Zeek package is located

        name (str): the canonical name of the package, which is always the
            last component of the git URL path

        source (str): the package source this package comes from, which
            may be empty if the package is not a part of a source (i.e. the user
            is referring directly to the package's git URL).

        directory (str): the directory within the package source where the
            :file:`zkg.index` containing this package is located.
            E.g. if the package source has a package named "foo" declared in
            :file:`alice/zkg.index`, then `dir` is equal to "alice".
            It may also be empty if the package is not part of a package source
            or if it's located in a top-level :file:`zkg.index` file.

        metadata (dict of str -> str): the contents of the package's
            :file:`zkg.meta` or :file:`bro-pkg.meta` file.  If the package has
            not been installed then this information may come from the last
            aggregation of the source's :file:`aggregate.meta` file (it may not
            be accurate/up-to-date).
    """

    def __init__(self, git_url, source='', directory='', metadata=None,
                 name=None, canonical=False):
        self.git_url = git_url
        self.source = source
        self.directory = directory
        self.metadata = {} if metadata is None else metadata
        self.name = name

        if not canonical:
            self.git_url = canonical_url(git_url)

            if not source and os.path.exists(git_url):
                # Ensures getting real path of relative directories.
                # e.g. canonical_url catches "./foo" but not "foo"
                self.git_url = os.path.realpath(self.git_url)

            self.name = name_from_path(git_url)

    def __str__(self):
        return self.qualified_name()

    def __repr__(self):
        return self.git_url

    def __lt__(self, other):
        return str(self) < str(other)

    def aliases(self):
        """Return a list of package name aliases.

        The canonical one is listed first.
        """
        return aliases(self.metadata)

    def tags(self):
        """Return a list of keyword tags associated with the package.

        This will be the contents of the package's `tags` field and may
        return results from the source's aggregated metadata if the package
        has not been installed yet."""
        return tags(self.metadata)

    def short_description(self):
        """Return a short description of the package.

        This will be the first sentence of the package's 'description' field
        and may return results from the source's aggregated metadata if the
        package has not been installed yet."""
        return short_description(self.metadata)

    def dependencies(self, field='depends'):
        """Returns a dictionary of dependency -> version strings.

        The keys indicate the name of a package (shorthand name or full git
        URL).  The names 'zeek' or 'zkg' may also be keys that indicate a
        dependency on a particular Zeek or zkg version.

        The values indicate a semantic version requirement.

        If the dependency field is malformed (e.g. number of keys not equal to
        number of values), then None is returned.
        """
        return dependencies(self.metadata, field)

    def user_vars(self):
        """Returns a list of (str, str, str) from metadata's 'user_vars' field.

        Each entry in the returned list is a the name of a variable, it's value,
        and its description.

        If the 'user_vars' field is not present, an empty list is returned.  If
        it is malformed, then None is returned.
        """
        return user_vars(self.metadata)

    def name_with_source_directory(self):
        """Return the package's name within its package source.

        E.g. for a package source with a package named "foo" in
        :file:`alice/zkg.index`, this method returns "alice/foo".
        If the package has no source or sub-directory within the source, then
        just the package name is returned.
        """
        if self.directory:
            return '{}/{}'.format(self.directory, self.name)

        return self.name

    def qualified_name(self):
        """Return the shortest name that qualifies/distinguishes the package.

        If the package is part of a source, then this returns
        "source_name/:meth:`name_with_source_directory()`", else the package's
        git URL is returned.
        """
        if self.source:
            return '{}/{}'.format(self.source,
                                  self.name_with_source_directory())

        return self.git_url

    def matches_path(self, path):
        """Return whether this package has a matching path/name.

        E.g for a package with :meth:`qualified_name()` of "zeek/alice/foo",
        the following inputs will match: "foo", "alice/foo", "zeek/alice/foo"
        """
        path_parts = path.split('/')

        if self.source:
            pkg_path = self.qualified_name()
            pkg_path_parts = pkg_path.split('/')

            for i, part in reversed(list(enumerate(path_parts))):
                ri = i - len(path_parts)

                if part != pkg_path_parts[ri]:
                    return False

            return True
        else:
            if len(path_parts) == 1 and path_parts[-1] == self.name:
                return True

            return path == self.git_url



from setuptools import setup

install_requires = ['gitpython', 'semantic_version', 'btest']

setup(
    name='zkg',
    version=open('VERSION').read().replace('-', '.dev', 1).strip(),
    description='The Zeek Package Manager',
    long_description=open('README').read(),
    license='University of Illinois/NCSA Open Source License',
    keywords='zeek zeekctl zeekcontrol package manager scripts plugins security',
    maintainer='The Zeek Project',
    maintainer_email='info@zeek.org',
    url='https://github.com/zeek/package-manager',
    scripts=['zkg'],
    packages=['zeekpkg'],
    install_requires=install_requires,
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'License :: OSI Approved :: University of Illinois/NCSA Open Source License',
        'Operating System :: POSIX :: Linux',
        'Operating System :: MacOS :: MacOS X',
        'Programming Language :: Python :: 3',
        'Topic :: System :: Networking :: Monitoring',
        'Topic :: Utilities',
    ],
)



"""
A module for zkg's notion of "user variables": named values required
by packages that the user can provide in a variety of ways, including
responses to zkg's input prompting.
"""
import os
import re
import readline

def slugify(string):
    """Returns file-system-safe, lower-case version of the input string.

    Any character sequence outside of ``[a-zA-Z0-9_]+`` gets replaced by a
    single underscore. If the variable has no value or the value is an
    empty string, returns the given default.
    """
    return re.sub(r'[^\w]+', '_', string, flags=re.ASCII).lower()


def _rlinput(prompt, prefill=''):
    """Variation of input() that supports pre-filling a value."""
    readline.set_startup_hook(lambda: readline.insert_text(prefill))
    try:
        return input(prompt)
    finally:
        readline.set_startup_hook()


class UserVar():
    """A class representing a single user variable.

    User variables have a name and an optional description. They
    resolve to a value using a cascade of mechanisms, including
    command-line arguments (via --user-var), environment variables,
    cached previous values, and user input. They may come with a
    default value.
    """
    def __init__(self, name, val=None, default=None, desc=None):
        self._name = name
        self._desc = desc or ''
        self._val = val
        self._default = default if default is not None else val

    def name(self):
        return self._name

    def desc(self):
        return self._desc

    def set(self, val):
        self._val = val

    def val(self, fallback=None):
        return self._val if self._val is not None else fallback

    def default(self):
        return self._default

    def resolve(self, name, config, user_var_args=None, force=False):
        """Populates user variables with updated values and returns them.

        This function resolves the variable in the following order:

        (1) Use any value provided on the command line via --user-var
        (2) If force is not used, prompt the user for input
        (3) use an environment variables of the same name,
        (4) retrieve from the provided config parser's "user_vars" section,
        (5) use the default value of the user variable.

        The resolved value is stored with the instance (to be
        retrieved via .val() in the future) and also returned.

        Args:
            name (str): the requesting entity, e.g. a package name

            config (configparser.ConfigParser): the zkg configuration

            user_var_args (list of UserVar): user-var instances
                provided via command line

            force (bool): whether to skip prompting for input

        Returns:
            str: the resulting variable value

        Raises:
            ValueError: when we couldn't produce a value for the variable
        """
        val = None
        source = None

        if user_var_args:
            for uvar in user_var_args:
                if uvar.name() == self._name:
                    val = uvar.val()
                    source = 'command line'
                    break

        if val is None:
            val = os.environ.get(self._name)
            if val:
                source = 'environment'

        if source:
            print('"{}" will use value of "{}" ({}) from {}: {}'.format(
                name, self._name, self._desc, source, val))
            self._val = val
            return val

        if val is None:
            # Try to re-use a cached value in the subsequent prompt
            val = config.get('user_vars', self._name, fallback=self._default)

        if force:
            if val is None:
                raise ValueError(self._name)
            self._val = val
            return val

        desc = ' (' + self._desc + ')' if self._desc else ''
        print('"{}" requires a "{}" value{}: '.format(
            name, self._name, desc))
        self._val = _rlinput(self._name + ': ', val)

        return self._val

    @staticmethod
    def parse_arg(arg):
        """Parser for NAME=VAL format string used in command-line args."""
        try:
            name, val = arg.split('=', 1)
            return UserVar(name, val=val)
        except ValueError as error:
            raise ValueError('invalid user var argument "{}", must be NAME=VAR'
                             .format(arg)) from error

    @staticmethod
    def parse_dict(metadata_dict):
        """Returns list of UserVars from the metadata's 'user_vars' field.

        Args:
            metadata_dict (dict of str->str): the input metadata, e.g. from
                a configparser entry value.

        Returns:
            list of UserVar. If the 'user_vars' field is not present,
            an empty list is returned.  If malformed, returns None.
        """
        text = metadata_dict.get('user_vars')

        if not text:
            return []

        rval = []

        text = text.strip()
        # Example: LIBRDKAFKA_ROOT [/usr] "Path to librdkafka installation"

        entries = re.split(r'(\w+\s+\[.*\]\s+".*")\s+', text)
        entries = list(filter(None, entries))

        for entry in entries:
            mob = re.match(r'(\w+)\s+\[(.*)\]\s+"(.*)"', entry)

            if not mob:
                return None

            groups = mob.groups()

            if len(groups) != 3:
                return None

            rval.append(UserVar(groups[0], val=groups[1], desc=groups[2]))

        return rval
